/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
  Copyright (C) 2012 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.

compile: 
g++ -O3 -I/home/spanderen/workspace/QuantLib -o fdheston fdheston.cpp -L/home/spanderen/workspace/QuantLib/ql/.libs -lQuantLib  -lboost_chrono -lboost_system

*/


#include <ql/quotes/simplequote.hpp>
#include <ql/time/daycounters/actual360.hpp>
#include <ql/instruments/vanillaoption.hpp>
#include <ql/models/equity/hestonmodel.hpp>
#include <ql/termstructures/yield/zerocurve.hpp>
#include <ql/termstructures/yield/flatforward.hpp>
#include <ql/pricingengines/vanilla/fdhestonvanillaengine.hpp>
#include <ql/math/richardsonextrapolation.hpp>

#include <boost/chrono/chrono_io.hpp>


using namespace QuantLib;
using namespace boost::chrono;

class FdRichExtraHestonVanillaEngine
    : public GenericModelEngine<HestonModel,
                                 DividendVanillaOption::arguments,
                                 DividendVanillaOption::results> {
  public:
    // Constructor
    FdRichExtraHestonVanillaEngine(
       const boost::shared_ptr<HestonModel>& model,
       const boost::shared_ptr<VanillaOption>& option,
       Real rateOfConvergence,
       Size tGrid = 100, Size xGrid = 100,
       Size vGrid = 50, Size dampingSteps = 0,
       const FdmSchemeDesc& schemeDesc = FdmSchemeDesc::Hundsdorfer())
    : GenericModelEngine<HestonModel,
                            DividendVanillaOption::arguments,
                            DividendVanillaOption::results>(model),
      option_(option),
      rateOfConvergence_(rateOfConvergence),
      tGrid_(tGrid), xGrid_(xGrid),
      vGrid_(vGrid), dampingSteps_(dampingSteps),
      schemeDesc_(schemeDesc) { }

    void calculate() const {
        const RichardsonExtrapolation richExtra(1, *this);
        if (rateOfConvergence_ > 0.0) {
            results_.value = richExtra.formula(2.0, rateOfConvergence_);
        }
        else {
            results_.value = richExtra.formulaUnknownConvergence(2.0, 4.0);
        }
    }

    Real operator()(Real t) {
        const boost::shared_ptr<PricingEngine> engine(
            new FdHestonVanillaEngine(model_.currentLink(),
                tGrid_/t, xGrid_, vGrid_, dampingSteps_, schemeDesc_));

        engine->reset();
        option_->setupArguments(engine->getArguments());
        engine->getArguments()->validate();
        engine->calculate();

        const Real npv = dynamic_cast<const DividendVanillaOption::results*>(
            engine->getResults())->value;

        return npv;
    }

  private:
    const boost::shared_ptr<VanillaOption> option_;
    const Real rateOfConvergence_;
    const Size tGrid_, xGrid_, vGrid_, dampingSteps_;
    const FdmSchemeDesc schemeDesc_;
};


boost::shared_ptr<YieldTermStructure>
flatRate(const Date& today, Rate forward, const DayCounter& dc) {
    return boost::shared_ptr<YieldTermStructure>(
        new FlatForward(today, Handle<Quote>(
            boost::shared_ptr<Quote>(new SimpleQuote(forward))), dc));
}

int main() {

    const Date todaysDate = Date(28, March, 2004);
    Settings::instance().evaluationDate() = todaysDate;

    Handle<YieldTermStructure> rTS(flatRate(todaysDate, 0.10, Actual360()));
    Handle<YieldTermStructure> qTS(flatRate(todaysDate, 0.0 , Actual360()));

    const Date exerciseDate(26, March, 2005);

    const boost::shared_ptr<Exercise> americanExercise(
        new AmericanExercise(exerciseDate));

    const boost::shared_ptr<StrikedTypePayoff> payoff(new
        PlainVanillaPayoff(Option::Put, 10));

    const boost::shared_ptr<VanillaOption> americanOption(
        new VanillaOption(payoff, americanExercise));

    const Handle<Quote> s0(boost::shared_ptr<Quote>(new SimpleQuote(11)));

    const boost::shared_ptr<HestonProcess> hestonProcess(
        new HestonProcess(rTS, qTS, s0, 0.0625, 5, 0.16, 0.9, 0.1));

    const boost::shared_ptr<HestonModel> hestonModel(
        new HestonModel(hestonProcess));

    Real strikes[]  = { 11 };

    for (Size t=2; t <= 32768; t*=2) {
        for (Size i=0; i < sizeof(strikes)/sizeof(Real); ++i) {
        
            const boost::shared_ptr<PricingEngine> rEngineOne(
                new FdRichExtraHestonVanillaEngine(
                    hestonModel, americanOption, 1.0, t, 400, 100));

            const boost::shared_ptr<PricingEngine> rEngineUnknown(
                new FdRichExtraHestonVanillaEngine(
                    hestonModel, americanOption, 0.0, t, 400, 100));

            const boost::shared_ptr<PricingEngine> fdEngine(
                 new FdHestonVanillaEngine(hestonModel, t, 400, 100));

            const boost::shared_ptr<PricingEngine> engines[] = {
                rEngineOne, rEngineUnknown, fdEngine };

            for (Size j=0; j < 3; ++j) {
                americanOption->setPricingEngine(engines[j]);

                high_resolution_clock::time_point t0 
                    = high_resolution_clock::now();
                Real calculated = americanOption->NPV();
            
                std::cout << std::setw(18) << std::setprecision(16)
                          << calculated << " "
                          << high_resolution_clock::now() - t0
                          << std::endl;
            }
        }
    }
}

